#include "App.h"
#include "../Framework3/Math/Noise.h"


BaseApp *app = new App();

void App::resetCamera(){
	camPos = vec3(0, 1, -6);
	wx = 0;
	wy = 0;
	speed = 4.0f;
}

vec3 *dest;

void subDivide(const vec3 &v0, const vec3 &v1, const vec3 &v2, int level){
	if (level){
		level--;
		vec3 v3 = normalize(v0 + v1);
		vec3 v4 = normalize(v1 + v2);
		vec3 v5 = normalize(v2 + v0);

		subDivide(v0, v3, v5, level);
		subDivide(v3, v4, v5, level);
		subDivide(v3, v1, v4, level);
		subDivide(v5, v4, v2, level);
	} else {
		*dest++ = v0;
		*dest++ = v1;
		*dest++ = v2;
	}
}

bool App::init(){
	currRT = 0;

	initNoise();

	int subDivLevel = 4;
	int nVertices = 4 * 3 * (1 << (2 * subDivLevel));

	vec3 *vertices = new vec3[nVertices];

	float e = 1.0f / (1 + sqrtf(2));
	float a = sqrtf(1 - e * e);
	float b = 1 + e - a;
	float d = sqrtf(1 - e * e - b * b);
	float3 y  = float3( 0,  1,  0);
	float3 x  = float3( a, -e,  0);
	float3 z0 = float3(-b, -e,  d);
	float3 z1 = float3(-b, -e, -d);

	dest = vertices;
	subDivide(y, z0, x,  subDivLevel);
	subDivide(y, z1, z0, subDivLevel);
	subDivide(y, x,  z1, subDivLevel);
	subDivide(x, z0, z1, subDivLevel);

	model = new Model();
	model->setIndexCount(nVertices);
	model->addStream(TYPE_VERTEX, 3, nVertices, (float *) vertices, model->getArrayIndices(nVertices), false);

	return true;
}

void App::exit(){
	delete model;
}

bool App::load(){
	if (d3d->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, 0, D3DRTYPE_SURFACE, (D3DFORMAT) MAKEFOURCC('R','2','V','B')) != D3D_OK){
		ErrorMsg("This demo requires R2VB support.");
		return false;
	}

	bool supportsHalf = false;
	if (d3d->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, 0, D3DRTYPE_TEXTURE, D3DFMT_A16B16G16R16F) == D3D_OK){
		if (caps.DeclTypes & D3DDTCAPS_FLOAT16_4) supportsHalf = true;
	}


	if (caps.PixelShaderVersion < D3DPS_VERSION(2,0)){
		ErrorMsg("This demo requires pixel shader 2.0.");
		return false;
	}

	if ((nearestClamp = renderer->addSamplerState(NEAREST, CLAMP, CLAMP, CLAMP)) == SS_NONE) return false;
	if ((linear = renderer->addSamplerState(LINEAR, WRAP, WRAP, WRAP)) == SS_NONE) return false;
	if ((bilinearClamp = renderer->addSamplerState(BILINEAR, CLAMP, CLAMP, CLAMP)) == SS_NONE) return false;
	if ((trilinearAniso = renderer->addSamplerState(TRILINEAR_ANISO, WRAP, WRAP, WRAP)) == SS_NONE) return false;


	if ((hairPosRT[0] = renderer->addRenderTarget(STRAND_LENGTH, STRAND_COUNT, FORMAT_RGBA16F, nearestClamp, R2VB)) == TEXTURE_NONE) return false;
	if ((hairDirRT[0] = renderer->addRenderTarget(STRAND_LENGTH, STRAND_COUNT, FORMAT_RGBA16F, nearestClamp, R2VB)) == TEXTURE_NONE) return false;
	if ((hairPosRT[1] = renderer->addRenderTarget(STRAND_LENGTH, STRAND_COUNT, FORMAT_RGBA16F, nearestClamp, R2VB)) == TEXTURE_NONE) return false;
	if ((hairDirRT[1] = renderer->addRenderTarget(STRAND_LENGTH, STRAND_COUNT, FORMAT_RGBA16F, nearestClamp, R2VB)) == TEXTURE_NONE) return false;

	if ((hairFinalRT  = renderer->addRenderTarget(STRAND_LENGTH, STRAND_COUNT, supportsHalf? FORMAT_RGBA16F : FORMAT_RGBA32F, nearestClamp, R2VB)) == TEXTURE_NONE) return false;


	// Shaders
	if ((initHair     = renderer->addShader("initHair.shd"    )) == SHADER_NONE) return false;
	if ((hairPhysics  = renderer->addShader("hairPhysics.shd" )) == SHADER_NONE) return false;
	if ((renderHair   = renderer->addShader("renderHair.shd"  )) == SHADER_NONE) return false;
	if ((renderHead   = renderer->addShader("renderHead.shd"  )) == SHADER_NONE) return false;
	if ((renderFloor  = renderer->addShader("renderFloor.shd" )) == SHADER_NONE) return false;
	if ((hairLighting = renderer->addShader("hairLighting.shd")) == SHADER_NONE) return false;

	if ((skyBox = renderer->addShader("skybox.shd")) == SHADER_NONE) return false;



	// Generate an index buffer for the hair
	uint *indices = new uint[INDEX_COUNT];
	uint *iDest = indices;
	for (int i = 0; i < STRAND_COUNT; i++){
		for (int k = 0; k < STRAND_LENGTH - 1; k++){
			*iDest++ = i * STRAND_LENGTH + k;
			*iDest++ = i * STRAND_LENGTH + k + 1;
		}
	}
	if ((hairIndices = renderer->addIndexBuffer(INDEX_COUNT, sizeof(uint), STATIC, indices)) == IB_NONE) return false;
	delete indices;



	// Detect whether the debug runtime is active
	bool debugRuntime = false;
	HKEY hKey;
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Direct3D", 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS){
		DWORD size = sizeof(DWORD);
		DWORD value;
		if (RegQueryValueEx(hKey, "LoadDebugRuntime", NULL, NULL, (LPBYTE) &value, &size) == ERROR_SUCCESS){
			debugRuntime = (value != 0);
		}
		RegCloseKey(hKey);
	}

	if ((dummy = renderer->addVertexBuffer(debugRuntime? STRAND_LENGTH * STRAND_COUNT * 4 * (supportsHalf? sizeof(half) : sizeof(float)) : 1, STATIC)) == SHADER_NONE) return false;



	Image img;
	half *dest = (half *) img.create(FORMAT_RGBA16F, STRAND_LENGTH, STRAND_COUNT, 1, 1);
	for (int s = 0; s < STRAND_COUNT; s++){
		float x, y, z;

		do {
			x = float(rand()) / RAND_MAX * 2.0f - 1.0f;
			z = float(rand()) / RAND_MAX;
			z = (1 - z * z) * 1.39f - 0.4f;

			y = 1.0f - x * x - z * z;
		} while (y < 0.0f);
		
		y = sqrtf(y);

		for (int i = 0; i < STRAND_LENGTH; i++){
			float f = (1 + i * 2.0f / (STRAND_LENGTH - 1.0f));

			*dest++ = 0.8f * f * x;
			*dest++ = y * f;
			*dest++ = 0.8f * f * z;
			*dest++ = 1.0f;
		}
	}
	if ((startHair = renderer->addTexture(img, nearestClamp)) == TEXTURE_NONE) return false;


	if ((floor = renderer->addNormalMap("../Textures/glass03.jpg", FORMAT_RGBA8, true, trilinearAniso)) == TEXTURE_NONE) return false;

	if ((head = renderer->addTexture("../Textures/face.dds", false, linear)) == TEXTURE_NONE) return false;


	// Cubemap
	const char *files[] = {
		"../Textures/CubeMaps/Giger-hive/posx.jpg", "../Textures/CubeMaps/Giger-hive/negx.jpg",
		"../Textures/CubeMaps/Giger-hive/posy.jpg", "../Textures/CubeMaps/Giger-hive/negy.jpg",
		"../Textures/CubeMaps/Giger-hive/posz.jpg", "../Textures/CubeMaps/Giger-hive/negz.jpg",
	};
	if ((env = renderer->addCubemap(files, true, bilinearClamp)) == TEXTURE_NONE) return false;


	// Vertex declaration
	FormatDesc qFormat[] = {
		0, TYPE_VERTEX, FORMAT_FLOAT, 2,
	};
	if ((quadVF = renderer->addVertexFormat(qFormat, elementsOf(qFormat))) == VF_NONE) return false;

	FormatDesc hFormat[] = {
		0, TYPE_VERTEX, supportsHalf? FORMAT_HALF : FORMAT_FLOAT, 4,
	};
	if ((hairVF = renderer->addVertexFormat(hFormat, elementsOf(hFormat))) == VF_NONE) return false;

	FormatDesc sbFormat[] = {
		0, TYPE_VERTEX,   FORMAT_FLOAT, 4,
		0, TYPE_TEXCOORD, FORMAT_FLOAT, 4,
	};
	if ((sbVF = renderer->addVertexFormat(sbFormat, elementsOf(sbFormat))) == VF_NONE) return false;


	if (!model->makeDrawable(renderer, true)) return false;

	return onReset();
}

void App::unload(){
	
}

bool App::onReset(){
	needsInit = true;

	return true;
}


void App::drawEnvironment(const float4x4 &mvp){
	renderer->reset();
	renderer->setShader(skyBox);
	renderer->setVertexFormat(sbVF);
	renderer->setTexture("Env", env, bilinearClamp);
	renderer->setDepthState(noDepthWrite);
	renderer->apply();

	float4x4 iMvp = !mvp;

	float4 v0( 0,  2, 1, 1);
	float4 v1( 3, -1, 1, 1);
	float4 v2(-3, -1, 1, 1);

	float4 vertices[] = {
		v0, iMvp * v0,
		v1, iMvp * v1,
		v2, iMvp * v2,
	};

	dev->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 1, vertices, 2 * sizeof(float4));
}

void App::drawFrame(){
	mat4 projection = toD3DProjection(perspectiveMatrixX(1.5f, width, height, 0.01f, 30));
	mat4 modelview = rotateXY(-wx, -wy);

	mat4 envMvp = projection * modelview;
	modelview.translate(-camPos);

	mat4 mvp = projection * modelview;


	float3 lightDir = normalize(float3(-0.3f, 0.5f, -1));


	// Setup head movement
	float y = fmodf(0.3f * time, 1.0f);
	y *= (1 - y);
	y *= (noise1(floorf(0.3f * time) * 0.28439f) * 4 + 4);

	static vec3 lastHeadPos[64], headPos;

	for (int i = 0; i < 63; i++){
		lastHeadPos[i] = lastHeadPos[i + 1];
	}
	lastHeadPos[63] = headPos;
	headPos = vec3(8 * noise1(2.1298f - 0.19582f * time), y + 1, 8 * noise1(0.9182f + 0.2142f * time));

	vec3 diff = headPos - lastHeadPos[0];
	float angle = atan2f(diff.x, -diff.z);

	mat4 headMV = translate(headPos) * rotateY(angle + noise1(0.423f * time));


	if (keys[KEY_F5]) needsInit = true;

	int otherRT = 1 - currRT;

	TextureID rt[] = { hairPosRT[currRT], hairDirRT[currRT] };
	renderer->changeRenderTargets(rt, elementsOf(rt));

		const float vertices[] = { -1, -1,   1, -1,   1, 1,   -1, 1 };
	
		if (needsInit){
			/*
				Initialize hair
			*/
			float2 scale = float2(STRAND_LENGTH / (STRAND_LENGTH - 1.0f), -STRAND_COUNT / (STRAND_COUNT - 1.0f));
			float2 bias = float2(1.0f / (STRAND_LENGTH - 1.0f), 1.0f / (STRAND_COUNT - 1.0f));

			renderer->reset();
			renderer->setShader(initHair);
			renderer->setTexture("Hair", startHair, nearestClamp);
			renderer->setVertexFormat(quadVF);
			renderer->setDepthState(noDepthTest);
			renderer->setShaderConstant2f("scale", scale);
			renderer->setShaderConstant2f("bias", bias);
			renderer->setShaderConstant4x4f("headMV", headMV);
			renderer->apply();

			dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(float2));

			needsInit = false;
		} else {
			/*
				Hair simulation pass
			*/
			float fTime = min(frameTime, 0.007f);

			renderer->reset();
			renderer->setShader(hairPhysics);
			renderer->setTexture("Hair", startHair, nearestClamp);
			renderer->setTexture("Pos", hairPosRT[otherRT], nearestClamp);
			renderer->setTexture("Dir", hairDirRT[otherRT], nearestClamp);
			renderer->setVertexFormat(quadVF);
			renderer->setDepthState(noDepthTest);

			renderer->setShaderConstant2f("halfPixel", float2(0.5f / STRAND_LENGTH, 0.5f / STRAND_COUNT));
			renderer->setShaderConstant2f("texel", float2(1.0f / STRAND_LENGTH, 0));
			renderer->setShaderConstant1f("dampen", powf(0.35f, fTime));
			renderer->setShaderConstant1f("frameTime", fTime);
			renderer->setShaderConstant1f("forceFactor", 4000.0f * fTime);
			renderer->setShaderConstant1f("naturalLength", 2.0f / (STRAND_LENGTH - 1.0f));

			vec3 windDir = 5.0f * float3(noise1(time), 0.4f * noise1(time * 0.7283f + 0.51982f), noise1(-time));
			windDir.y -= 3.25f; // Gravity
			renderer->setShaderConstant3f("windDir", fTime * windDir);

			renderer->setShaderConstant4x4f("headMV", headMV);
			renderer->setShaderConstant3f("headPos", headPos);
			renderer->apply();


			dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(float2));
		}

	renderer->changeRenderTarget(hairFinalRT);
		/*
			Compute lighting and write to .w
		*/
		renderer->reset();
		renderer->setShader(hairLighting);
		renderer->setTexture("Pos", hairPosRT[currRT], nearestClamp);
		renderer->setVertexFormat(quadVF);
		renderer->setDepthState(noDepthTest);
		renderer->setShaderConstant2f("halfPixel", float2(0.5f / STRAND_LENGTH, 0.5f / STRAND_COUNT));
		renderer->setShaderConstant2f("offset", float2(1.0f / STRAND_LENGTH, 0.0));
		renderer->setShaderConstant3f("lightDir", lightDir);
		renderer->apply();

		dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(float2));
	
	renderer->changeToMainFramebuffer();







	dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);

	/*
		Render the head
	*/
	renderer->reset();
	renderer->setShader(renderHead);
	renderer->setTexture("Head", head, linear);
	renderer->setRasterizerState(cullBack);

	renderer->setShaderConstant4x4f("mvp", mvp * headMV);

	float4 lDir = (!headMV) * float4(lightDir, 0.0f);

	renderer->setShaderConstant3f("lightDir", normalize(lDir.xyz()));
	renderer->apply();

	model->draw(renderer);





	/*
		Render the floor
	*/
	renderer->reset();
	renderer->setShader(renderFloor);
	renderer->setTexture("Floor", floor, trilinearAniso);
	renderer->setVertexFormat(quadVF);
	renderer->setShaderConstant4x4f("mvp", mvp * translate(0, -0.01f, 0) * scale(8, 0, 8));
	renderer->setShaderConstant3f("lightDir", float3(lightDir.x, lightDir.z, lightDir.y));
	renderer->apply();

	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(float2));

	



	/*
		Render the hair
	*/
	renderer->reset();
	renderer->setShader(renderHair);
	renderer->setVertexFormat(hairVF);
	renderer->setVertexBuffer(0, dummy); // Dummy
	renderer->setIndexBuffer(hairIndices);
	renderer->setShaderConstant4x4f("mvp", mvp);
	renderer->apply();

	dev->SetRenderState(D3DRS_POINTSIZE, R2VB_ENABLE);
	dev->SetRenderState(D3DRS_POINTSIZE, R2VB_MAP_STREAM(0, R2VB_DMAP));
	dev->SetTexture(D3DDMAPSAMPLER, ((Direct3DRenderer *) renderer)->getD3DTexture(hairFinalRT));

	dev->DrawIndexedPrimitive(D3DPT_LINELIST, 0, 0, STRAND_COUNT * STRAND_LENGTH, 0, INDEX_COUNT / 2);

	dev->SetTexture(D3DDMAPSAMPLER, NULL);
	dev->SetRenderState(D3DRS_POINTSIZE, R2VB_UNMAP_STREAM(0));
	dev->SetRenderState(D3DRS_POINTSIZE, R2VB_DISABLE);


	drawEnvironment(envMvp);


	currRT = otherRT;
}
